@@ -25,6 +25,9 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#ifdef G_OS_WIN32
+#include <io.h>
+#endif
 
 #include <string.h>
 #include <fcntl.h>
@@ -38,11 +41,20 @@
 
 #define DEFAULT_SECTION _("Others")
 
+/* Properties */
+enum {
+	PROP_0,
+	PROP_ID,
+	PROP_NAME,
+	PROP_SECTION,
+	PROP_HIDDEN
+};
+
 G_DEFINE_TYPE (GtkSourceLanguage, gtk_source_language, G_TYPE_OBJECT)
 
-static GtkSourceLanguage *process_language_node 		(xmlTextReaderPtr 		 reader,
-								 const gchar 			*filename);
-
+static GtkSourceLanguage *process_language_node (xmlTextReaderPtr	 reader,
+						 const gchar		*filename);
+static gboolean		  force_styles		(GtkSourceLanguage	*language);
 
 GtkSourceLanguage *
 _gtk_source_language_new_from_file (const gchar              *filename,
@@ -107,12 +119,48 @@ _gtk_source_language_new_from_file (const gchar              *filename,
 	{
 		lang->priv->language_manager = lm;
 		g_object_add_weak_pointer (G_OBJECT (lm),
-					   (gpointer*) &lang->priv->language_manager);
+					   (gpointer) &lang->priv->language_manager);
 	}
 
 	return lang;
 }
 
+static void
+gtk_source_language_get_property (GObject    *object,
+				  guint       prop_id,
+				  GValue     *value,
+				  GParamSpec *pspec)
+{
+	GtkSourceLanguage *language;
+
+	g_return_if_fail (GTK_IS_SOURCE_LANGUAGE (object));
+
+	language = GTK_SOURCE_LANGUAGE (object);
+
+	switch (prop_id)
+	{
+		case PROP_ID:
+			g_value_set_string (value, language->priv->id);
+			break;
+
+		case PROP_NAME:
+			g_value_set_string (value, language->priv->name);
+			break;
+
+		case PROP_SECTION:
+			g_value_set_string (value, language->priv->name);
+			break;
+
+		case PROP_HIDDEN:
+			g_value_set_boolean (value, language->priv->hidden);
+			break;
+
+		default:
+			G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
+			break;
+	}
+}
+
 static void
 gtk_source_language_dispose (GObject *object)
 {
@@ -123,11 +171,11 @@ gtk_source_language_dispose (GObject *object)
 	if (lang->priv->language_manager != NULL)
 	{
 		g_object_remove_weak_pointer (G_OBJECT (lang->priv->language_manager),
-					      (gpointer*) &lang->priv->language_manager);
+					      (gpointer) &lang->priv->language_manager);
 		lang->priv->language_manager = NULL;
 	}
 
-	G_OBJECT_CLASS (gtk_source_language_parent_class)->finalize (object);
+	G_OBJECT_CLASS (gtk_source_language_parent_class)->dispose (object);
 }
 
 static void
@@ -146,6 +194,7 @@ gtk_source_language_finalize (GObject *object)
 	g_free (lang->priv->section);
 	g_free (lang->priv->id);
 	g_hash_table_destroy (lang->priv->properties);
+
 	g_hash_table_destroy (lang->priv->styles);
 
 	G_OBJECT_CLASS (gtk_source_language_parent_class)->finalize (object);
@@ -156,8 +205,41 @@ gtk_source_language_class_init (GtkSourceLanguageClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-	object_class->dispose	= gtk_source_language_dispose;
-	object_class->finalize	= gtk_source_language_finalize;
+	object_class->get_property = gtk_source_language_get_property;
+	object_class->dispose = gtk_source_language_dispose;
+	object_class->finalize = gtk_source_language_finalize;
+
+	g_object_class_install_property (object_class,
+					 PROP_ID,
+					 g_param_spec_string ("id",
+						 	      _("Language id"),
+							      _("Language id"),
+							      NULL,
+							      G_PARAM_READABLE));
+
+	g_object_class_install_property (object_class,
+					 PROP_ID,
+					 g_param_spec_string ("name",
+						 	      _("Language name"),
+							      _("Language name"),
+							      NULL,
+							      G_PARAM_READABLE));
+
+	g_object_class_install_property (object_class,
+					 PROP_ID,
+					 g_param_spec_string ("section",
+						 	      _("Language section"),
+							      _("Language section"),
+							      NULL,
+							      G_PARAM_READABLE));
+
+	g_object_class_install_property (object_class,
+					 PROP_HIDDEN,
+					 g_param_spec_boolean ("hidden",
+							       _("Hidden"),
+							       _("Whether the language should be hidden from the user"),
+							       FALSE,
+							       G_PARAM_READABLE));
 
 	g_type_class_add_private (object_class, sizeof(GtkSourceLanguagePrivate));
 }
@@ -167,7 +249,10 @@ gtk_source_language_init (GtkSourceLanguage *lang)
 {
 	lang->priv = G_TYPE_INSTANCE_GET_PRIVATE (lang, GTK_TYPE_SOURCE_LANGUAGE,
 						  GtkSourceLanguagePrivate);
-	lang->priv->styles = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
+	lang->priv->styles = g_hash_table_new_full (g_str_hash,
+						    g_str_equal,
+						    g_free,
+						    (GDestroyNotify)_gtk_source_style_info_free);
 	lang->priv->properties = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_free);
 }
 
@@ -190,11 +275,10 @@ static void
 process_properties (xmlTextReaderPtr   reader,
 		    GtkSourceLanguage *language)
 {
-	gint ret;
 	xmlNodePtr child;
 	xmlNodePtr node = NULL;
 
-	while (node == NULL && (ret = xmlTextReaderRead (reader)) == 1)
+	while (node == NULL && xmlTextReaderRead (reader) == 1)
 	{
 		xmlChar *name;
 
@@ -433,6 +517,22 @@ gtk_source_language_get_section	(GtkSourceLanguage *language)
 	return language->priv->section;
 }
 
+/**
+ * gtk_source_language_get_hidden:
+ * @language: a #GtkSourceLanguage
+ *
+ * Returns whether the language should be hidden from the user.
+ *
+ * Returns: TRUE if the language should be hidden, FALSE otherwise.
+ */
+gboolean
+gtk_source_language_get_hidden (GtkSourceLanguage *language)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (language), FALSE);
+
+	return language->priv->hidden;
+}
+
 /**
  * gtk_source_language_get_metadata:
  * @language: a #GtkSourceLanguage.
@@ -455,7 +555,6 @@ gtk_source_language_get_metadata (GtkSourceLanguage *language,
 
 /**
  * gtk_source_language_get_mime_types:
- *
  * @language: a #GtkSourceLanguage.
  *
  * Returns the mime types associated to this language. This is just
@@ -483,7 +582,6 @@ gtk_source_language_get_mime_types (GtkSourceLanguage *language)
 
 /**
  * gtk_source_language_get_globs:
- *
  * @language: a #GtkSourceLanguage.
  *
  * Returns the globs associated to this language. This is just
@@ -525,32 +623,68 @@ _gtk_source_language_get_language_manager (GtkSourceLanguage *language)
 
 /* Highlighting engine creation ------------------------------------------ */
 
+static void
+copy_style_info (const char         *style_id,
+		 GtkSourceStyleInfo *info,
+		 GHashTable         *dest)
+{
+	g_hash_table_insert (dest, g_strdup (style_id),
+			     _gtk_source_style_info_copy (info));
+}
+
 void
 _gtk_source_language_define_language_styles (GtkSourceLanguage *lang)
 {
-#define ADD_ALIAS(style,mapto)								\
-	g_hash_table_insert (lang->priv->styles, g_strdup (style), g_strdup (mapto))
+	static const gchar *alias[][2] = {
+		{"Base-N Integer", "def:base-n-integer"},
+		{"Character", "def:character"},
+		{"Comment", "def:comment"},
+		{"Function", "def:function"},
+		{"Decimal", "def:decimal"},
+		{"Floating Point", "def:floating-point"},
+		{"Keyword", "def:keyword"},
+		{"Preprocessor", "def:preprocessor"},
+		{"String", "def:string"},
+		{"Specials", "def:specials"},
+		{"Data Type", "def:type"},
+		{NULL, NULL}};
 
-	ADD_ALIAS ("Base-N Integer", "def:base-n-integer");
-	ADD_ALIAS ("Character", "def:character");
-	ADD_ALIAS ("Comment", "def:comment");
-	ADD_ALIAS ("Function", "def:function");
-	ADD_ALIAS ("Decimal", "def:decimal");
-	ADD_ALIAS ("Floating Point", "def:floating-point");
-	ADD_ALIAS ("Keyword", "def:keyword");
-	ADD_ALIAS ("Preprocessor", "def:preprocessor");
-	ADD_ALIAS ("String", "def:string");
-	ADD_ALIAS ("Specials", "def:specials");
-	ADD_ALIAS ("Data Type", "def:data-type");
+	gint i = 0;
+	GtkSourceLanguageManager *lm;
+	GtkSourceLanguage *def_lang;
 
-#undef ADD_ALIAS
+	while (alias[i][0] != NULL)
+	{
+		GtkSourceStyleInfo *info;
+
+		info = _gtk_source_style_info_new (alias[i][0], alias[i][1]);
+
+		g_hash_table_insert (lang->priv->styles,
+				     g_strdup (alias[i][0]),
+				     info);
+
+		++i;
+	}
+
+	/* We translate String to def:string, but def:string is mapped-to
+	 * def:constant in def.lang, so we got to take style mappings from def.lang */
+
+	lm = _gtk_source_language_get_language_manager (lang);
+	def_lang = gtk_source_language_manager_get_language (lm, "def");
+
+	if (def_lang != NULL)
+	{
+		force_styles (def_lang);
+		g_hash_table_foreach (def_lang->priv->styles,
+				      (GHFunc) copy_style_info,
+				      lang->priv->styles);
+	}
 }
 
-GtkSourceEngine *
-_gtk_source_language_create_engine (GtkSourceLanguage *language)
+/* returns new reference, which _must_ be unref'ed */
+static GtkSourceContextData *
+gtk_source_language_parse_file (GtkSourceLanguage *language)
 {
-	GtkSourceContextEngine *ce = NULL;
-
 	if (language->priv->ctx_data == NULL)
 	{
 		gboolean success = FALSE;
@@ -587,11 +721,193 @@ _gtk_source_language_create_engine (GtkSourceLanguage *language)
 		_gtk_source_context_data_ref (language->priv->ctx_data);
 	}
 
-	if (language->priv->ctx_data)
+	return language->priv->ctx_data;
+}
+
+GtkSourceEngine *
+_gtk_source_language_create_engine (GtkSourceLanguage *language)
+{
+	GtkSourceContextEngine *ce = NULL;
+	GtkSourceContextData *ctx_data;
+
+	ctx_data = gtk_source_language_parse_file (language);
+
+	if (ctx_data != NULL)
 	{
-		ce = _gtk_source_context_engine_new (language->priv->ctx_data);
-		_gtk_source_context_data_unref (language->priv->ctx_data);
+		ce = _gtk_source_context_engine_new (ctx_data);
+		_gtk_source_context_data_unref (ctx_data);
 	}
 
 	return ce ? GTK_SOURCE_ENGINE (ce) : NULL;
 }
+
+typedef struct _AddStyleIdData AddStyleIdData;
+
+struct _AddStyleIdData
+{
+	gchar             *language_id;
+	GPtrArray         *ids_array;
+};
+
+static void
+add_style_id (gchar *id, G_GNUC_UNUSED gpointer value, AddStyleIdData *data)
+{
+	if (g_str_has_prefix (id, data->language_id))
+		g_ptr_array_add (data->ids_array, g_strdup (id));
+}
+
+static gchar **
+get_style_ids (GtkSourceLanguage *language)
+{
+	GPtrArray *ids_array;
+	AddStyleIdData data;
+
+	g_return_val_if_fail (language->priv->styles != NULL, NULL);
+
+	ids_array = g_ptr_array_new ();
+
+	data.language_id = g_strdup_printf ("%s:", language->priv->id);
+	data.ids_array = ids_array;
+
+	g_hash_table_foreach (language->priv->styles,
+			      (GHFunc) add_style_id,
+			      &data);
+
+	g_free (data.language_id);
+
+	if (ids_array->len == 0)
+	{
+		/* No style defined in this language */
+		g_ptr_array_free (ids_array, TRUE);
+
+		return NULL;
+	}
+	else
+	{
+		/* Terminate the array with NULL */
+		g_ptr_array_add (ids_array, NULL);
+
+		return (gchar **)g_ptr_array_free (ids_array, FALSE);
+	}
+}
+
+static gboolean
+force_styles (GtkSourceLanguage *language)
+{
+	/* To be sure to have the list of styles we need to parse lang file
+	 * as if we were to create an engine. In the future we can improve
+	 * this by parsing styles only.
+	 */
+	if (!language->priv->styles_loaded && language->priv->ctx_data == NULL)
+	{
+		GtkSourceContextData *ctx_data;
+
+		ctx_data = gtk_source_language_parse_file (language);
+		if (ctx_data == NULL)
+			return FALSE;
+
+		language->priv->styles_loaded = TRUE;
+		_gtk_source_context_data_unref (ctx_data);
+	}
+
+	return TRUE;
+}
+
+/**
+ * gtk_source_language_get_style_ids:
+ * @language: a #GtkSourceLanguage
+ *
+ * Returns the ids of the styles defined by this @language.
+ *
+ * Returns: a  %NULL terminated array containing
+ * ids of the styles defined by this @language or %NULL if no style is
+ * defined.  The returned array must be freed with g_strfreev().
+*/
+gchar **
+gtk_source_language_get_style_ids (GtkSourceLanguage *language)
+{
+	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (language), NULL);
+	g_return_val_if_fail (language->priv->id != NULL, NULL);
+
+	if (!force_styles (language))
+		return NULL;
+
+	return get_style_ids (language);
+}
+
+static GtkSourceStyleInfo *
+get_style_info (GtkSourceLanguage *language, const char *style_id)
+{
+	GtkSourceStyleInfo *info;
+
+	if (!force_styles (language))
+		return NULL;
+
+	g_return_val_if_fail (language->priv->styles != NULL, NULL);
+
+	info = g_hash_table_lookup (language->priv->styles, style_id);
+
+	return info;
+}
+
+/**
+ * gtk_source_language_get_style_name:
+ * @language: a #GtkSourceLanguage
+ * @style_id: a style ID
+ *
+ * Returns the name of the style with ID @style_id defined by this @language.
+ *
+ * Returns: the name of the style with ID @style_id defined by this @language or
+ * %NULL if the style has no name or there is no style with ID @style_id defined
+ * by this @language. The returned string is owned by the @language and must
+ * not be modified.
+ */
+const char *
+gtk_source_language_get_style_name (GtkSourceLanguage *language,
+				    const char        *style_id)
+{
+	GtkSourceStyleInfo *info;
+
+	g_return_val_if_fail (GTK_IS_SOURCE_LANGUAGE (language), NULL);
+	g_return_val_if_fail (language->priv->id != NULL, NULL);
+	g_return_val_if_fail (style_id != NULL, NULL);
+
+	info = get_style_info (language, style_id);
+	if (info == NULL)
+		return NULL;
+
+	return info->name;
+}
+
+/* Utility functions for GtkSourceStyleInfo */
+
+GtkSourceStyleInfo *
+_gtk_source_style_info_new (const gchar *name, const gchar *map_to)
+{
+	GtkSourceStyleInfo *info = g_new0 (GtkSourceStyleInfo, 1);
+
+	info->name = g_strdup (name);
+	info->map_to = g_strdup (map_to);
+
+	return info;
+}
+
+GtkSourceStyleInfo *
+_gtk_source_style_info_copy (GtkSourceStyleInfo *info)
+{
+	g_return_val_if_fail (info != NULL, NULL);
+	return _gtk_source_style_info_new (info->name, info->map_to);
+}
+
+void
+_gtk_source_style_info_free (GtkSourceStyleInfo *info)
+{
+	if (info == NULL)
+		return;
+
+	g_free (info->name);
+	g_free (info->map_to);
+
+	g_free (info);
+}
+
